package mahjong;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import mahjong.automata.DFA;
import mahjong.automata.DFAEquivalenceChecker;
import mahjong.automata.DFAFactory;
import mahjong.automata.DFAState;
import mahjong.fpg.FieldPointstoGraph;
import mahjong.pta.Obj;
import mahjong.pta.Type;
import mahjong.util.UnionFindSet;

/**
 * @author Tian Tan
 * @author Yue Li
 */
public class HeapAbstraction {

	private final FieldPointstoGraph fpg;
	private final DFAFactory dfaFactory;
	private final DFAEquivalenceChecker dfaEqChecker;
	
	/** This map would be manipulated by multiple threads
	 *  thus it should be concurrent hashmap. */
	private Map<Obj, Boolean> canMerged;
	
	public HeapAbstraction(FieldPointstoGraph fpg) {
		this.fpg = fpg;
		this.dfaFactory = new DFAFactory(fpg);
		this.dfaEqChecker = new DFAEquivalenceChecker();
	}
	
	public Map<Obj, Obj> computeMergedObjectMap() {
		UnionFindSet<Obj> uf = modelHeap();
		Map<Obj, Obj> mom = convertToMap(uf.getDisjointSets());
		return mom;
	}
	
	/**
	 * Modeling the heap by checking the equivalence of automata.
	 */
	private UnionFindSet<Obj> modelHeap() {
		canMerged = new ConcurrentHashMap<>();
		Set<Obj> allObjs = fpg.getAllObjs();
		UnionFindSet<Obj> uf = new UnionFindSet<>(allObjs);
		// group the objects by their types
		Map<Type, Set<Obj>> groupedObjs = allObjs
				.stream()
				.collect(Collectors.groupingBy(
						obj -> obj.getType(), Collectors.toSet()));
		groupedObjs.entrySet()
				.parallelStream()
				.forEach(entry -> {
					Set<Obj> objs = entry.getValue();
					DFAMap dfaMap = new DFAMap();
					for (Obj o1 : objs) {
						if (canBeMerged(o1, dfaMap)) {
							for (Obj o2 : objs) {
								if (canBeMerged(o2, dfaMap)) {
									if (o1.getID() <= o2.getID()
											&& !uf.isConnected(o1, o2)) {
										if (canBeMerged(o1, o2, dfaMap)) {
											uf.union(o1, o2);
										}
									}
								}
							}
						}
					}
				});
		return uf;
	}
	
	/**
	 * 
	 * @param o1
	 * @param o2
	 * @return whether o1 and o2 can be merged.
	 */
	private boolean canBeMerged(Obj o1, Obj o2, DFAMap dfaMap) {
		if (o1 == o2) { return true; }
		DFA dfa1 = dfaMap.retrieveDFA(o1);
		DFA dfa2 = dfaMap.retrieveDFA(o2);
		return dfaEqChecker.isEquivalent(dfa1, dfa2);
	}
	
	/**
	 * 
	 * @param o
	 * @return whether o can be merged with other objects.
	 */
	private boolean canBeMerged(Obj o, DFAMap dfaMap) {
		if (!canMerged.containsKey(o)) {
			boolean result = true;
			// Check whether the types of objects pointed (directly/indirectly)
			// by o are unique.
			DFA dfa = dfaMap.retrieveDFA(o);
			for (DFAState s : dfa.getStates()) {
				if (dfa.outputOf(s).size() > 1) {
					// Types pointed (directly/indirectly) by o are not unique.
					result = false;
					break;
				}
			}
			canMerged.put(o, result);
		}
		return canMerged.get(o);
	}
	
	private static Map<Obj, Obj> convertToMap(Collection<Set<Obj>> objModel) {
		Map<Obj, Obj> map = new HashMap<>();
		objModel.forEach(objs -> {
			Obj rep = selectRepresentative(objs);
			objs.forEach(obj -> map.put(obj, rep));
		});
		return map;
	}
	
	private static Obj selectRepresentative(Set<Obj> objs) {
		return objs.stream().findFirst().get();
	}
    
    /**
     * During equivalence check, each thread holds a DFAMap which contains
     * the DFA of the objects of the type. After comparison, the DFAMap
     * and its containing DFA will be released to save memory space.
     *
     */
    private class DFAMap {
        
        private final Map<Obj, DFA> dfaMap = new HashMap<>();
        
        private DFA retrieveDFA(Obj o) {
            if (!dfaMap.containsKey(o)) {
                DFA dfa = dfaFactory.getDFA(o);
                dfaMap.put(o, dfa);
            }
            return dfaMap.get(o);
        }
    }

}
