package mahjong.automata;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import mahjong.pta.Field;
import mahjong.pta.Type;

/**
 * @author Tian Tan
 * @author Yue Li
 */
public class DFA {
	
	private Set<DFAState> states, allStates;
	private DFAState q0;
	
	private static final DFAState deadState
			= new DFAState(Collections.emptySet(),
					Collections.emptySet());
	
	public DFA(DFAState q0) {
		this.q0 = q0;
	}
	
	/**
	 * 
	 * @return Set of states. Does not contains dead state.
	 */
	public Set<DFAState> getStates() {
		if (states == null) {
			computeStates();
		}
		return states;
	}
	
	/**
	 * 
	 * @return Set of all states including dead state.
	 */
	public Set<DFAState> getAllStates() {
		if (allStates == null) {
			computeStates();
		}
		return allStates;
	}
	
	private void computeStates() {
		Queue<DFAState> queue = new LinkedList<>();
		queue.add(q0);
		states = new HashSet<>();
		while (!queue.isEmpty()) {
			DFAState s = queue.poll();
			if (!states.contains(s)) {
				states.add(s);
				s.getNextMap().values().forEach(queue::add);
			}
		}
		allStates = new HashSet<>(states);
		allStates.add(deadState);
	}
	
	public DFAState getStartState() {
		return q0;
	}
	
	public DFAState getDeadState() {
		return deadState;
	}
	
	public boolean isDeadState(DFAState s) {
		return deadState == s;
	}
	
	public DFAState nextState(DFAState s, Field f) {
		if (isDeadState(s)) {
			return getDeadState();
		}
		Map<Field, DFAState> nextMap = s.getNextMap();
		if (nextMap.containsKey(f)) {
			return nextMap.get(f);
		}
		return getDeadState();
	}
	
	public Set<Type> outputOf(DFAState s) {
		return s.getOutput();
	}
	
	public Set<Field> outEdgesOf(DFAState s) {
		Map<Field, DFAState> nextMap = s.getNextMap();
		return nextMap.keySet();
	}
	
	public boolean containsState(DFAState s) {
		return getAllStates().contains(s);
	}
	
}
