package mahjong.automata;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.stream.Collectors;


import mahjong.fpg.FieldPointstoGraph;
import mahjong.pta.Field;
import mahjong.pta.Obj;
import mahjong.pta.Type;

/**
 * @author Tian Tan
 * @author Yue Li
 */
public class DFAFactory {

	private final FieldPointstoGraph fpg;
	
	private Map<Set<Obj>, DFAState> stateMap;
	private Set<DFAState> states, visited;
	
	public DFAFactory(FieldPointstoGraph fpg) {
		this.fpg = fpg;
		buildAllDFA();
	}
	
	public DFA getDFA(Obj obj) {
		DFAState q0 = stateMap.get(Collections.singleton(obj));
		return new DFA(q0);
	}
	
	private void buildAllDFA() {
		stateMap = new HashMap<>();
		states = new HashSet<>();
		visited = new HashSet<>();
		fpg.getAllObjs().forEach(this::buildDFA);
	}
	
	/**
	 * Perform subset construction algorithm to convert an NFA
	 * to a DFA. If a set of NFA states are merged to an existing
	 * DFA state, then reused the existing DFA state instead of creating
	 * an equivalent new one.
	 * @param obj the start state (object) of the DFA
	 */
	private void buildDFA(Obj obj) {
		Set<Obj> q0Set = Collections.singleton(obj);
		if (!stateMap.containsKey(q0Set)) {
			NFA nfa = new NFA(obj, fpg);
			DFAState startState = getDFAState(q0Set, nfa);
			Queue<DFAState> worklist = new LinkedList<>();
			states.add(startState);
			worklist.add(startState);
			while (!worklist.isEmpty()) {
				DFAState s = worklist.poll();
				if (!visited.contains(s)) {
					visited.add(s);
					Set<Field> fields = fields(nfa, s.getObjects());
					fields.forEach(f -> {
						Set<Obj> nextNFAStates = move(nfa, s.getObjects(), f);
						DFAState nextState = getDFAState(nextNFAStates, nfa);
						if (!states.contains(nextState)) {
							states.add(nextState);
							worklist.add(nextState);
						}
						addTransition(s, f, nextState);
					});
				}
			}
		}
	}
	
	private DFAState getDFAState(Set<Obj> objs, NFA nfa) {
		if (!stateMap.containsKey(objs)) {
			Set<Type> output = objs.stream()
					.map(nfa::outputOf)
					.collect(Collectors.toSet());
			stateMap.put(objs, new DFAState(objs, output));
		}
		return stateMap.get(objs);
	}
	
	private Set<Obj> move(NFA nfa, Set<Obj> objs, Field f) {
		return objs.stream()
				.map(obj -> nfa.nextStates(obj, f))
				.flatMap(nextStates -> nextStates.stream())
				.collect(Collectors.toSet());
	}
	
	private Set<Field> fields(NFA nfa, Set<Obj> objs) {
		return objs.stream()
				.map(nfa::outEdgesOf)
				.flatMap(fields -> fields.stream())
				.collect(Collectors.toSet());
	}
	
	private void addTransition(DFAState s, Field f, DFAState nextState) {
		s.addTransition(f, nextState);
	}
	
}
