package mahjong.fpg;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import mahjong.pta.Field;
import mahjong.pta.Obj;
import mahjong.pta.PTAProvider;

/**
 * @author Tian Tan
 * @author Yue Li
 */
public class FieldPointstoGraph {

	private PTAProvider provider;
	private Map<Obj, Map<Field, Set<Obj>>> pointsTo = new HashMap<>();
	private Map<Obj, Map<Field, Set<Obj>>> pointedBy = new HashMap<>();
	
	public FieldPointstoGraph(PTAProvider provider) {
		this.provider = provider;
		provider.objIterator().forEachRemaining(this::insertObj);
		provider.fptIterator().forEachRemaining(triple -> {
			Obj baseObj = triple.getFirst();
			Field field = triple.getSecond();
			Obj obj = triple.getThird();
			insertFPT(baseObj, field, obj);
		});
	}
	
	public PTAProvider getPTAProvider() {
		return provider;
	}
	
	public Set<Obj> getAllObjs() {
		return pointsTo.keySet();
	}
	
	public Set<Field> outFieldsOf(Obj baseObj) {
		return pointsTo.get(baseObj).keySet();
	}
	
	public Set<Field> inFieldsOf(Obj obj) {
		return pointedBy.get(obj).keySet();
	}
	
	public Set<Obj> pointsTo(Obj baseObj, Field field) {
		return pointsTo.get(baseObj).get(field);
	}
	
	public Set<Obj> pointedBy(Obj obj, Field field) {
		return pointedBy.get(obj).get(field);
	}
	
	public boolean hasFieldPointer(Obj obj, Field field) {
		return pointsTo.get(obj).containsKey(field);
	}
	
	
	private void insertObj(Obj obj) {
		if (!pointsTo.containsKey(obj)) {
			pointsTo.put(obj, new HashMap<>());
		}
		if (!pointedBy.containsKey(obj)) {
			pointedBy.put(obj, new HashMap<>());
		}
	}
	
	/**
	 * Insert field points-to relation.
	 * @param baseObj the base object
	 * @param field   a field of `baseObj'
	 * @param obj     the object pointed by `field'
	 */
	private void insertFPT(Obj baseObj, Field field, Obj obj) {
		insertPointsTo(baseObj, field, obj);
		insertPointedBy(baseObj, field, obj);
	}
	
	private void insertPointsTo(Obj baseObj, Field field, Obj obj) {
		Map<Field, Set<Obj>> fpt = pointsTo.get(baseObj);
		if (!fpt.containsKey(field)) {
			fpt.put(field, new HashSet<>());
		}
		fpt.get(field).add(obj);
	}
	
	private void insertPointedBy(Obj baseObj, Field field, Obj obj) {
		Map<Field, Set<Obj>> fpb = pointedBy.get(obj);
		if (!fpb.containsKey(field)) {
			fpb.put(field, new HashSet<>());
		}
		fpb.get(field).add(baseObj);
	}
	
}
