package mahjong.main;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.Map;

import mahjong.HeapAbstraction;
import mahjong.fpg.FieldPointstoGraph;
import mahjong.pta.Obj;
import mahjong.pta.doop.DataBase;
import mahjong.pta.doop.DoopPTAProvider;
import mahjong.pta.doop.Options;

/**
 * @author Tian Tan
 * @author Yue Li
 */
public class DoopMain {

	private static final char SEP = '\t';
	private static final char EOL = '\n';
	
	public static void main(String[] args) throws FileNotFoundException {
		run(args);
	}

	public static void run(String[] args) throws FileNotFoundException {
		Options opt = Options.parse(args);
		
		FieldPointstoGraph fpg = buildFPG(opt.getDbPath());
		
		System.out.print("Creating heap abstraction ... ");
		long start = System.currentTimeMillis();
		HeapAbstraction heapAbs = new HeapAbstraction(fpg);
		Map<Obj, Obj> mom = heapAbs.computeMergedObjectMap();
		long end = System.currentTimeMillis();
		DoopMain.outputElapsedTime(start, end);
		
		outputStatistics(mom);
		System.out.printf("Writing Mahjong heap abstraction to %s ...\n", opt.getOutPath());
		File outputFile = new File(opt.getOutPath(), "MahjongHeapAbstraction.facts");
		writeMergedObjectMap(mom, outputFile);
	}
	
	public static FieldPointstoGraph buildFPG(String dbPath) {
		File dbDir = new File(dbPath);
		DataBase db = new DataBase(dbDir);
		
		System.out.print("Building FPG (Field Points-to Graph) ... ");
		long start = System.currentTimeMillis();
		DoopPTAProvider fptProvider = new DoopPTAProvider(db);
		FieldPointstoGraph fpg = new FieldPointstoGraph(fptProvider);
		long end = System.currentTimeMillis();
		outputElapsedTime(start, end);
		
		return fpg;
	}
	
	public static FieldPointstoGraph buildFPG(String dbPath, String cachePath,
			String app) {
		File dbDir = new File(dbPath);
		File cacheDir = new File(cachePath);
		DataBase db = new DataBase(dbDir, cacheDir, app);	
		
		System.out.print("Building FPG (Field Points-to Graph) ... ");
		long start = System.currentTimeMillis();
		DoopPTAProvider fptProvider = new DoopPTAProvider(db);
		FieldPointstoGraph fpg = new FieldPointstoGraph(fptProvider);
		long end = System.currentTimeMillis();
		outputElapsedTime(start, end);
		
		return fpg;
	}
	
	public static void outputElapsedTime(long start, long end) {
		System.out.printf("elapsed time: %.2fs\n", (end - start) / 1000F);
	}
	
	public static void outputStatistics(Map<Obj, Obj> mom) {
		int nObj = (int) mom.keySet().stream().distinct().count();
		int nObjMahjong = (int) mom.values().stream().distinct().count();
		System.out.println("-----------------------------------------------------------");
		System.out.printf("%d objects in the allocation-site heap abstraction.\n", nObj);
		System.out.printf("%d objects in the Mahjong heap abstraction.\n", nObjMahjong);
		System.out.println("-----------------------------------------------------------");
	}
	
	private static void writeMergedObjectMap(Map<Obj, Obj> mom,
			File outputFile) throws FileNotFoundException {
		PrintWriter writer = new PrintWriter(outputFile);
		mom.forEach((heap, mergedHeap) -> {
			writer.write(heap.getName());
			writer.write(SEP);
			writer.write(mergedHeap.getName());
			writer.write(EOL);
		});
		writer.close();
	}
	
}
