package mahjong.pta.doop;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * The class representing database of Doop.
 *
 * @author Tian Tan
 * @author Yue Li
 */
public class DataBase {

	public static final Map<String, String> QUERY = new HashMap<>();
	private static final String CMD = "bloxbatch -db %s -query %s";
	private static final String SEP = ", ";
	
	private Map<String, List<List<String>>> queryResults = new HashMap<>();
	
	private final File dbDir;
	private final File cacheDir;
	private final String program;
	
	static {
		QUERY.put("OBJ", "_(obj)<-Stats:Simple:InsensVarPointsTo(obj,_).");
		QUERY.put("OBJ_TYPE", "_[obj]=type<-"
				+ "HeapAllocation:Type[obj]=type,"
				+ "Stats:Simple:InsensVarPointsTo(obj,_).");
		QUERY.put("IFPT", "_(baseheap,field,heap)<-"
				+ "InstanceFieldPointsTo(_,heap,field,_,baseheap).");
		QUERY.put("APT", "_(array,heap)<-ArrayIndexPointsTo(_,heap,_,array).");
	}
	
	
	public DataBase(File dbDir) {
		this.dbDir = null;
		this.cacheDir = null;
		this.program = null;
		init(dbDir);
	}
	
	public DataBase(File dbDir, File cacheDir, String program) {
		this.dbDir = dbDir;
		this.cacheDir = cacheDir;
		this.program = program;
	}
	
	/** Return the results of the give query. */
	List<List<String>> query(String query) {
		if (this.cacheDir == null) {
			return queryResults.get(query);
		} else {
			return queryDoopWithCache(query);
		}
	}
	
	/**
	 * Query Doop database and store the query results to queryResults. 
	 * @param dbDir the directory of Doop database.
	 */
	private void init(File dbDir) {
		System.out.println("Querying Doop database ...");
		for (String query: QUERY.keySet()) {
			queryDoop(dbDir, query);
		}
	}
	
	private void queryDoop(File dbDir, String query) {
		List<List<String>> results = new LinkedList<>();
		try {
			Process proc = null;
			String cmd = String.format(CMD, dbDir.getAbsolutePath(),
					QUERY.get(query));
			Runtime rt = Runtime.getRuntime();
			proc = rt.exec(cmd);
			BufferedReader reader = new BufferedReader(new InputStreamReader(
					proc.getInputStream()));
		
			String line = null;
			while ((line = reader.readLine()) != null) {
				results.add(line2list(line));
			}
			if (proc != null) {
				proc.waitFor();		
			}
			reader.close();
			queryResults.put(query, results);
		} catch (InterruptedException e) {
			throw new RuntimeException("Exception during query: "
					+ QUERY.get(query));
		} catch (IOException e) {
			throw new RuntimeException("Query " + query + " fails, " +
					"caused by " + e.getMessage());
		}
	}
	
	/** Query Doop and cache the results for later uses. */
	private List<List<String>> queryDoopWithCache(String query) {
		BufferedReader reader = getCacheReader(query);
		List<List<String>> result = new LinkedList<>();
		try {
			Process proc = null;
			if (reader == null) {
				String cmd = String.format(CMD, dbDir.getAbsolutePath(),
						QUERY.get(query));
				Runtime rt = Runtime.getRuntime();
				proc = rt.exec(cmd);
				reader = new BufferedReader(new InputStreamReader(
						proc.getInputStream()));
			}

			String line = null;
			while ((line = reader.readLine()) != null) {
				result.add(line2list(line));
			}
			if (proc != null) {
				proc.waitFor();
				// bloxbatch is invoked which means that the cache file
				// does not exist. So we are going to create the cache.
				createCache(query, result);				
			}
			reader.close();
		} catch (InterruptedException e) {
			throw new RuntimeException("Exception during query: "
					+ QUERY.get(query));
		} catch (IOException e) {
			throw new RuntimeException("Query " + query + " fails, " +
					"caused by " + e.getMessage());
		}
		return result;
	}
	
	private BufferedReader getCacheReader(String query) {
		BufferedReader reader = null;
		File path = getCacheFile(query);
		if (path.exists()) {
			try {
				reader = new BufferedReader(new FileReader(path));
			} catch (IOException e) {
				throw new RuntimeException("Read cache file fails: " + path);
			}
		}
		return reader;
	}
	
	private File getCacheFile(String query) {
		File cacheFile = null;
		if (cacheDir != null) {
			String fileName = String.format("%s.%s", program, query);
			String filePath = String.format("%s%s%s",
					cacheDir.getAbsolutePath(), File.separator, fileName);
			cacheFile = new File(filePath);
		}
		return cacheFile;
	}
	
	private void createCache(String query, List<List<String>> content) throws IOException {
		File path = getCacheFile(query);
		PrintWriter writer = new PrintWriter(new FileWriter(path));
		content.forEach(list -> {
			writer.println(String.join(SEP, list));
		});
		writer.close();
	}
	
	private List<String> line2list(String line) {
		return Arrays.asList(line.trim().split(SEP));
	}
	
}
